/* heston.C: implementation for analytic solution of HESTON's Model
   for stochastic volatility
   Copyright (C) 2001 Gunter Winkler (guwi17@gmx.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
  author: Gunter Winkler (gw) 
          gunter.winkler@mathematik.tu-chemnitz.de
 
  paper:  'A closed-form solution for options with stochastic volatility'
          by Heston, L., 1995 Volatility/Risk Publications
 
  changes:
 	gw	2001-08-07	new
 	gw	2001-08-23	made public first time
	gw  2002-08-10  changed integration rule to Gauss-Laguerre,
	                which is much faster
 */

#ifndef NDEBUG
#ifndef DEBUGLEVEL
#define DEBUGLEVEL 6
#endif
#endif

#include <cassert>
#include <cstdlib>
#include "hestonanalytic.h"

#define Pi  3.14159265358979323846264338328
#define Pi2 1.57079632679489661923132169164

/* names: phi = 1: call, phi = -1: put */

/* some defines */
#define _2U(j)	((j==1)?(1):(-1))
#define _A  	(market.kappa * market.theta)
#define _B(j) 	((j==1)?(market.kappa+market.lambda-market.rho*market.xi):(market.kappa+market.lambda))

static heston_complex Fj( const int j,	/* j == 1 or j != 1 */
						  const DOUBLE x, 
						  const DOUBLE v,
						  const DOUBLE tau,
						  const DOUBLE phi,
						  const MARKETDATA market)
/*
	helper function f_j(x,v,tau;phi) from paper
*/
{
	heston_complex	C, D;	/* for C(tau,phi) and D(tau,phi) */
	heston_complex	g, d;	/* see paper */
	heston_complex  z1, z2; /* temporary */
	DOUBLE			rpx;	/* = rho * phi * xi */
	DOUBLE			xi2;	/* = xi^2 */

	DEBUG_HESTON(9,{std::cout << "F_j(x=" << x << ", v=" << v << ", tau=" << tau << ", phi=" << phi << ")" << std::endl;});
		
	rpx = market.rho * market.xi * phi;
	xi2 = market.xi * market.xi;
	z1 = heston_complex( _B(j), -rpx );
	
	DEBUG_HESTON(10, std::cout << "rpx " << rpx << " xi^2 " << xi2 << " z1 " << z1 << std::endl;);

	DEBUG_HESTON(10, std::cout << "z1^2 " << (z1*z1) << " ... " << (xi2*(heston_complex(-phi*phi, _2U(j) * phi))););

	d = z1*z1 - xi2 * heston_complex(-phi*phi, _2U(j) * phi);
	DEBUG_HESTON(10, std::cout << " d " << d;);
	d = sqrt( d );
	
	DEBUG_HESTON(10, std::cout << " d " << d;);
	
	z2 = (z1 + d);
	g = z2 / (z1 - d);
	
	DEBUG_HESTON(10, std::cout << " z2 " << z2;);
	DEBUG_HESTON(9, std::cout << " g " << g;);

	D = z2 / xi2 * ((DOUBLE)1.0-exp(d*tau)) / ((DOUBLE)1.0 - g*exp(d*tau));
	DEBUG_HESTON(9, std::cout << " D " << D;);
		
	C = heston_complex((DOUBLE)0.0,(market.rd-market.rf)*phi*tau)
			+ _A/xi2 * ( z2 * tau - (DOUBLE)2.0 * log(((DOUBLE)1.0-g*exp(d*tau))/((DOUBLE)1.0-g)) );
	DEBUG_HESTON(9, std::cout << " C " << C << std::endl;);
	
	return exp( C + D * v + heston_complex(0.0, phi*x) );
}


double	Pj ( const int j,
			 const DOUBLE x,
			 const DOUBLE v,
			 const DOUBLE tau,
			 const DOUBLE logstrike,
			 MARKETDATA market,
			 p_intdata pdata)
/*
	conditional probability P_j(x,v,tau) from paper
*/
{
	const int       glsteps = 50; /* <= 150 for double precision */
	
	heston_complex	integrand;
	DOUBLE			phi, sum;
	int 			i;

	DEBUG_HESTON(8, std::cout << "P_j(x=" << x << ", v=" << v << ", tau=" << tau << ")" << std::endl;);
		
	/* Gauss-Laguerre */
	assert(NULL!=pdata);
	if (0==pdata->N) { /* compute coefficients */
	  double *sub;      /* subdiagonal */
	  pdata->N = glsteps;
	  pdata->w = (double*) calloc(sizeof(double), glsteps);
	  pdata->x = (double*) calloc(sizeof(double), glsteps);
	  sub = (double*) calloc(sizeof(double), glsteps);
	  assert(NULL!=pdata->w);
	  assert(NULL!=pdata->x);
	  assert(NULL!=sub);
	  for (i=1; i<glsteps; ++i) pdata->w[i]=0.0;
	  pdata->w[0]=1.0; 
	  for (i=0; i<glsteps; ++i) pdata->x[i]=2*i+1;
	  for (i=0; i<glsteps; ++i) sub[i]=i;
	  i = eig_tridi( 1, glsteps, pdata->x, sub, pdata->w );
	  DEBUG_HESTON(5, {std::cout << "eig_tridi returns " << i << std::endl;};);
	  assert(0==i);

	  for (i=0; i<glsteps; ++i) pdata->w[i] *= pdata->w[i] * exp(pdata->x[i]);

	  DEBUG_HESTON(8, {for (i=0; i<glsteps; ++i) {std::cout << "x=" << pdata->x[i] << ", " << "w=" << pdata->w[i] << std::endl;};}; );
	  free(sub);
	}
	assert(0 != pdata->N);
	assert(NULL != pdata->w);
	assert(NULL != pdata->x);

	sum = 0;
	for (i=glsteps-1;i>=0;--i)
	  {
		phi = pdata->x[i];
		integrand = (heston_complex) std::polar(1.0/phi,-phi*logstrike) 
		  * (heston_complex(0,-1.0))
		  * Fj(j,x,v,tau,phi,market);
		sum += pdata->w[i] * integrand.real();
		DEBUG_HESTON(9, std::cout << " phi " << phi << " int " << integrand << " sum " << sum << std::endl;);
	  }
	return 0.5 + sum / Pi;
}



double	Heston_Call( const DOUBLE spot,
					 const DOUBLE v,
					 const DOUBLE tau,
					 const DOUBLE strike,
					 MARKETDATA market,
					 p_intdata *pdata = NULL )
/*
	formula for vanilla call (same as Black-Scholes)
*/
{
	DOUBLE	x, ls, p1, p2;

	/* assure that memory for integration parameters is OK. */
	p_intdata pd = NULL;
	if (NULL!=pdata) {
	  if (NULL==*pdata) *pdata = (p_intdata) calloc(sizeof(intdata), 1);
	  pd = *pdata;
	}
	int freepdata = (NULL==pd);
	if (NULL==pd) pd = (p_intdata) calloc(sizeof(intdata), 1);
	assert(NULL!=pd);
	
	x = log(spot);
	ls = log(strike);

	DEBUG_HESTON(8,std::cout << "Heston_Call( x=" << x << ", v=" << v << ", tau=" << tau << ", K=" << strike << ")" << std::endl;);
	
	p1 = Pj(1, x, v, tau, ls, market, pd);
	p2 = Pj(2, x, v, tau, ls, market, pd);
	
	DEBUG_HESTON(8,std::cout << " P_1 " << p1 << ", P_2 " << p2 << std::endl;);
	
	if (freepdata) free(pd);

	return spot * exp(-market.rf * tau) * p1 - strike * exp(-market.rd * tau) * p2;

}
